"""Conditional Event item definition."""

from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.basicevent import draw_basic_event
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.ConditionalEvent)
class ConditionalEventItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=70, height=35)

        self.watch("subject.name").watch("subject.namespace.name")

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_basic_event,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("Conditional Event")]),
            text_name(self),
            text_from_package(self),
        )



"""Basic Event item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, ellipse, stroke
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.constants import DEFAULT_FTA_MAJOR
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.BasicEvent)
class BasicEventItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=DEFAULT_FTA_MAJOR, height=DEFAULT_FTA_MAJOR)

        self.watch("subject.name").watch("subject.namespace.name")

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_basic_event,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("Basic Event")]),
            text_name(self),
            text_from_package(self),
        )


def draw_basic_event(box, context: DrawContext, bounding_box: Rectangle):
    cr = context.cairo
    cr.move_to(bounding_box.width, bounding_box.height)
    ellipse(cr, *bounding_box)
    stroke(context, fill=True)



"""
ExtensionItem -- Graphical representation of an association.
"""

from gaphor import UML
from gaphor.diagram.presentation import LinePresentation, Named, text_name
from gaphor.diagram.shapes import Box
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.Extension)
class ExtensionItem(Named, LinePresentation):
    """ExtensionItem represents associations.

    An ExtensionItem has two ExtensionEnd items. Each ExtensionEnd item
    represents a Property (with Property.association == my association).
    """

    def __init__(self, diagram, id=None):
        super().__init__(
            diagram,
            id,
            shape_middle=Box(
                text_stereotypes(self),
                text_name(self),
            ),
        )

        self.watch("subject.name")
        self.watch("subject.appliedStereotype.classifier.name")

    def draw_head(self, context):
        cr = context.cairo
        cr.move_to(0, 0)
        cr.line_to(15, -10)
        cr.line_to(15, 10)
        cr.line_to(0, 0)
        cr.fill_preserve()
        cr.move_to(15, 0)



"""Format() and parse() provide a simple interface to:

1. format a model element to text
2. Update model element contents based on text
"""

from functools import singledispatch

from gaphor.core.modeling import Dependency, Diagram, Element, Relationship
from gaphor.i18n import gettext


@singledispatch
def format(el: Element, **kwargs) -> str:
    """Format an element."""
    raise TypeError(f"Format routine for type {type(el)} not implemented yet")


@singledispatch
def parse(el: Element, text: str) -> None:
    """Parse text and update `el` accordingly."""
    raise TypeError(f"Parsing routine for type {type(el)} not implemented yet")


@format.register(Element)
def format_namedelement(el: Element, **kwargs):
    return el.name or ""


@format.register(Diagram)
def format_diagram(el: Diagram, **kwargs) -> str:
    if el.diagramType:
        return f"[{el.diagramType}] {el.name}"
    return el.name or ""


@parse.register(Element)
def parse_namedelement(el: Element, text: str) -> None:
    """Parse element by simply assigning text to its name."""
    el.name = text


@parse.register(Diagram)
def parse_Diagram(el: Diagram, text: str) -> None:
    el.name = text


@format.register(Relationship)
def format_relationship(el):
    return el.__class__.__name__


@format.register(Dependency)
def format_dependency(el):
    return gettext("supplier: {name}").format(
        name=el.supplier and el.supplier.name or ""
    )



"""The painter module provides different painters for parts of the diagram.

Painters can be swapped in and out.

Each painter takes care of a layer in the diagram (such as grid, items
and handles).
"""

from __future__ import annotations

from cairo import LINE_JOIN_ROUND
from gaphas.geometry import Rectangle

from gaphor.core.modeling.diagram import DrawContext, StyledDiagram, StyledItem
from gaphor.diagram.diagramlabel import diagram_label
from gaphor.diagram.selection import Selection
from gaphor.diagram.shapes import Box, CssNode, Orientation, Text, cairo_state, stroke
from gaphor.UML.recipes import get_applied_stereotypes, stereotypes_str


class ItemPainter:
    def __init__(
        self, selection: Selection | None = None, dark_mode: bool | None = None
    ):
        self.selection: Selection = selection or Selection()
        self.dark_mode = dark_mode

    def paint_item(self, item, cr):
        selection = self.selection
        if not (diagram := item.diagram):
            return

        style = diagram.style(StyledItem(item, selection, self.dark_mode))

        cr.save()
        try:
            cr.set_line_join(LINE_JOIN_ROUND)
            cr.set_source_rgba(*style["color"])
            cr.transform(item.matrix_i2c.to_cairo())

            item.draw(
                DrawContext(
                    cairo=cr,
                    style=style,
                    selected=(item in selection.selected_items),
                    focused=(item is selection.focused_item),
                    hovered=(item is selection.hovered_item),
                    dropzone=(item is selection.dropzone_item),
                )
            )

        finally:
            cr.restore()

    def paint(self, items, cr):
        """Draw the items."""
        for item in items:
            self.paint_item(item, cr)


class DiagramTypePainter:
    def __init__(self, diagram):
        self.diagram = diagram
        self._pentagon = CssNode(
            "pentagon",
            diagram,
            Box(
                CssNode("diagramtype", None, Text(text=lambda: diagram.diagramType)),
                CssNode(
                    "stereotypes", None, Text(text=lambda: stereotypes_str(diagram))
                ),
                CssNode("name", None, Text(text=lambda: diagram_label(diagram))),
                orientation=Orientation.HORIZONTAL,
                draw=draw_pentagon,
            ),
        )

    def paint(self, _items, cr):
        diagram = self.diagram
        if not diagram.diagramType and not any(get_applied_stereotypes(diagram)):
            return

        style = diagram.style(StyledDiagram(diagram))

        context = DrawContext(
            cairo=cr,
            style=style,
            selected=False,
            focused=False,
            hovered=False,
            dropzone=False,
        )

        with cairo_state(cr):
            cr.identity_matrix()

            size = self._pentagon.size(context)  # type: ignore[arg-type]
            bb = Rectangle(0, 0, *size)
            self._pentagon.draw(context, bb)


def draw_pentagon(box, context, bounding_box):
    cr = context.cairo
    w = bounding_box.width
    h = bounding_box.height
    line_width = context.style.get("line-width", 2)
    with cairo_state(cr):
        h2 = h / 2.0
        cr.move_to(-line_width, -line_width)
        cr.line_to(-line_width, h)
        cr.line_to(w - 4, h)
        cr.line_to(w, h2)
        cr.line_to(w, -line_width)
        stroke(context, fill=True)



from __future__ import annotations

from gaphor import UML
from gaphor.core.modeling import Diagram
from gaphor.diagram.drop import drop, drop_relationship


@drop.register(UML.Association, Diagram)
def drop_association(element: UML.Association, diagram: Diagram, x, y):
    return drop_relationship(
        element, element.memberEnd[0].type, element.memberEnd[1].type, diagram, x, y
    )


@drop.register(UML.Connector, Diagram)
def drop_connector(element: UML.Connector, diagram: Diagram, x, y):
    return drop_relationship(
        element, element.end[0].role, element.end[1].role, diagram, x, y
    )


@drop.register(UML.Message, Diagram)
def drop_message(element: UML.Message, diagram: Diagram, x, y):
    return drop_relationship(
        element,
        element.sendEvent.covered
        if isinstance(element.sendEvent, UML.MessageOccurrenceSpecification)
        else None,
        element.receiveEvent.covered
        if isinstance(element.receiveEvent, UML.MessageOccurrenceSpecification)
        else None,
        diagram,
        x,
        y,
    )



from __future__ import annotations

from gaphor.abc import ActionProvider, Service
from gaphor.core import action
from gaphor.diagram.presentation import ElementPresentation
from gaphor.i18n import gettext
from gaphor.transaction import Transaction


class AlignService(Service, ActionProvider):
    def __init__(self, event_manager, diagrams, tools_menu=None):
        self.event_manager = event_manager
        self.diagrams = diagrams
        if tools_menu:
            tools_menu.add_actions(self)

    def shutdown(self):
        pass

    @action(name="align-left", label=gettext("Align left"))
    def align_left(self):
        self._modify_elements(_align_elements_left)

    @action(name="align-right", label=gettext("Align right"))
    def align_right(self):
        self._modify_elements(_align_elements_right)

    @action(name="align-vertical-center", label=gettext("Align vertical center"))
    def align_vertical_center(self):
        self._modify_elements(_align_elements_vertical_center)

    @action(name="align-top", label=gettext("Align top"))
    def align_top(self):
        self._modify_elements(_align_elements_top)

    @action(name="align-bottom", label=gettext("Align bottom"))
    def align_bottom(self):
        self._modify_elements(_align_elements_bottom)

    @action(name="align-horizontal-center", label=gettext("Align horizontal center"))
    def align_horizontal_center(self):
        self._modify_elements(_align_elements_horizontal_center)

    @action(name="resize-max-width", label=gettext("Max width"))
    def resize_max_width(self):
        self._modify_elements(_resize_elements_max_width)

    @action(name="resize-max-height", label=gettext("Max height"))
    def resize_max_height(self):
        self._modify_elements(_resize_elements_max_height)

    @action(name="resize-max-size", label=gettext("Max size"))
    def resize_max_size(self):
        self._modify_elements(_resize_elements_max_size)

    def _modify_elements(self, f):
        if current_view := self.diagrams.get_current_view():
            current_diagram = self.diagrams.get_current_diagram()
            selected_items = current_view.selection.selected_items

            if len(selected_items) >= 2:
                elements = {
                    item
                    for item in selected_items
                    if isinstance(item, ElementPresentation)
                }

                with Transaction(self.event_manager):
                    f(elements)
                    current_diagram.update(current_diagram.ownedPresentation)


def _align_elements_left(elements: set[ElementPresentation]):
    left_edge = _left_edge(elements)

    for item in elements:
        item.matrix.translate(left_edge - _pos_x(item), 0)


def _align_elements_right(elements: set[ElementPresentation]):
    right_edge = _right_edge(elements)

    for item in elements:
        item.matrix.translate(right_edge - (_pos_x(item) + item.width), 0)


def _align_elements_vertical_center(elements: set[ElementPresentation]):
    left_edge = _left_edge(elements)
    right_edge = _right_edge(elements)
    center_edge = left_edge + (right_edge - left_edge) / 2

    for item in elements:
        item.matrix.translate(center_edge - _pos_x(item) - item.width / 2, 0)


def _align_elements_top(elements: set[ElementPresentation]):
    top_edge = _top_edge(elements)

    for item in elements:
        item.matrix.translate(0, top_edge - _pos_y(item))


def _align_elements_bottom(elements: set[ElementPresentation]):
    bottom_edge = _bottom_edge(elements)

    for item in elements:
        item.matrix.translate(0, bottom_edge - (_pos_y(item) + item.height))


def _align_elements_horizontal_center(elements: set[ElementPresentation]):
    top_edge = _top_edge(elements)
    bottom_edge = _bottom_edge(elements)
    center_edge = top_edge + (bottom_edge - top_edge) / 2

    for item in elements:
        item.matrix.translate(0, center_edge - _pos_y(item) - item.height / 2)


def _resize_elements_max_width(elements: set[ElementPresentation]):
    max_width = _max_width(elements)

    for item in elements:
        item.width = max_width


def _resize_elements_max_height(elements: set[ElementPresentation]):
    max_height = _max_height(elements)

    for item in elements:
        item.height = max_height


def _resize_elements_max_size(elements: set[ElementPresentation]):
    max_width = _max_width(elements)
    max_height = _max_height(elements)

    for item in elements:
        item.width = max_width
        item.height = max_height


def _left_edge(elements: set[ElementPresentation]):
    return min({_pos_x(item) for item in elements})


def _right_edge(elements: set[ElementPresentation]):
    return max({_pos_x(item) + item.width for item in elements})


def _top_edge(elements: set[ElementPresentation]):
    return min({_pos_y(item) for item in elements})


def _bottom_edge(elements: set[ElementPresentation]):
    return max({_pos_y(item) + item.height for item in elements})


def _max_width(elements: set[ElementPresentation]):
    return max({item.width for item in elements})


def _max_height(elements: set[ElementPresentation]):
    return max({item.height for item in elements})


def _pos_x(item: ElementPresentation):
    _, _, _, _, x, _ = item.matrix.tuple()
    return x


def _pos_y(item: ElementPresentation):
    _, _, _, _, _, y = item.matrix.tuple()
    return y



"""NOT gate item definition."""

from gaphas.geometry import Rectangle

from gaphor.core.modeling import DrawContext
from gaphor.diagram.presentation import (
    Classified,
    ElementPresentation,
    text_name,
)
from gaphor.diagram.shapes import Box, IconBox, stroke
from gaphor.diagram.support import represents
from gaphor.RAAML import raaml
from gaphor.RAAML.fta.constants import DEFAULT_FTA_MAJOR, DEFAULT_FTA_MINOR
from gaphor.UML.compartments import text_from_package, text_stereotypes


@represents(raaml.NOT)
class NOTItem(Classified, ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, width=DEFAULT_FTA_MINOR, height=DEFAULT_FTA_MAJOR)

        self.watch("subject.name").watch("subject.namespace.name")

    def update_shapes(self, event=None):
        self.shape = IconBox(
            Box(
                draw=draw_not_gate,
            ),
            text_stereotypes(self, lambda: [self.diagram.gettext("NOT Gate")]),
            text_name(self),
            text_from_package(self),
        )


def draw_not_gate(box, context: DrawContext, bounding_box: Rectangle):
    cr = context.cairo
    x1 = 0
    x2 = bounding_box.width
    height = bounding_box.height
    cr.move_to(x1, height)
    cr.line_to(x2, 0)
    cr.line_to(x1, 0)
    cr.line_to(x2, height)
    cr.line_to(x1, height)
    stroke(context, fill=True)



"""Use case diagram item."""

from gaphor import UML
from gaphor.diagram.presentation import Classified, ElementPresentation, text_name
from gaphor.diagram.shapes import Box, draw_ellipse
from gaphor.diagram.support import represents
from gaphor.UML.compartments import text_stereotypes


@represents(UML.UseCase)
class UseCaseItem(Classified, ElementPresentation):
    """Presentation of gaphor.UML.UseCase."""

    def __init__(self, diagram, id=None):
        super().__init__(diagram, id, height=30)

        self.watch("subject.name")
        self.watch("subject.appliedStereotype.classifier.name")
        self.watch("subject[Classifier].isAbstract", self.update_shapes)

    def update_shapes(self, event=None):
        self.shape = Box(
            text_stereotypes(self),
            text_name(self),
            draw=draw_ellipse,
        )



"""CommentItem diagram item."""


from gaphor.core.modeling import Comment
from gaphor.diagram.presentation import ElementPresentation
from gaphor.diagram.shapes import Box, CssNode, Text, stroke
from gaphor.diagram.support import represents


@represents(Comment)
class CommentItem(ElementPresentation):
    def __init__(self, diagram, id=None):
        super().__init__(diagram, id)
        self.height = 50
        self.width = 100

        self.shape = Box(
            CssNode(
                "body",
                None,
                Text(
                    text=lambda: self.subject.body or "",
                ),
            ),
            draw=draw_border,
        )
        self.watch("subject[Comment].body")


def draw_border(box, context, bounding_box):
    cr = context.cairo
    x, y, w, h = bounding_box
    _, padding_right, _, padding_left = context.style.get("padding", (4, 16, 4, 4))
    ear = max(padding_right - padding_left, padding_left)
    line_to = cr.line_to
    cr.move_to(w - ear, y)
    line_to(w - ear, y + ear)
    line_to(w, y + ear)
    line_to(w - ear, y)
    line_to(x, y)
    line_to(x, h)
    line_to(w, h)
    line_to(w, y + ear)
    stroke(context, fill=True)



from gi.repository import Gtk

from gaphor.abc import ActionProvider
from gaphor.core import event_handler
from gaphor.core.modeling import ModelReady
from gaphor.event import ModelChangedOnDisk, ModelSaved
from gaphor.i18n import translated_ui_string
from gaphor.ui.abc import UIComponent


def new_builder():
    builder = Gtk.Builder()
    builder.add_from_string(translated_ui_string("gaphor.ui", "modelchanged.ui"))
    return builder


class ModelChanged(UIComponent, ActionProvider):
    def __init__(self, event_manager):
        self.event_manager = event_manager
        self._banner = None
        self._monitor = None
        self.event_manager.subscribe(self._on_show_banner)
        self.event_manager.subscribe(self._on_model_reset)

    def shutdown(self):
        super().shutdown()
        self.event_manager.subscribe(self._on_show_banner)
        self.event_manager.unsubscribe(self._on_model_reset)

    def open(self):
        """Open the diagrams component."""
        builder = new_builder()
        self._banner = builder.get_object("model-changed")
        return self._banner

    def close(self):
        """Close the diagrams component."""
        self._banner = None

    @event_handler(ModelChangedOnDisk)
    def _on_show_banner(self, _event):
        if self._banner:
            self._banner.set_revealed(True)

    @event_handler(ModelReady, ModelSaved)
    def _on_model_reset(self, _event):
        if self._banner:
            self._banner.set_revealed(False)



# ruff: noqa: F401,F403

from gaphor.core.modeling.coremodel import (
    Comment,
    Dependency,
    ElementChange,
    PendingChange,
    Picture,
    RefChange,
    Relationship,
    ValueChange,
)
from gaphor.core.modeling.diagram import (
    Diagram,
    DrawContext,
    UpdateContext,
)
from gaphor.core.modeling.element import Element, self_and_owners
from gaphor.core.modeling.elementfactory import ElementFactory
from gaphor.core.modeling.event import *
from gaphor.core.modeling.presentation import Presentation
from gaphor.core.modeling.stylesheet import StyleSheet
